#include <stdlib.h>
#include <string.h>
#include <write12.h>
#ifdef _WANT_SYSLOG
# include <syslog.h>
#endif
#include <fcntl.h>
#include "klumpat.h"

enum { TYPE_NULL = 0, TYPE_CONSOLE, TYPE_FILE, TYPE_META
#ifdef _WANT_SYSLOG
  ,TYPE_SYSLOG 
#endif
};

static char* level_table[] = { "DEBUG", "INFO", "NOTICE", "WARNING", "ERROR", "CRITICAL", "ALERT", "EMERGENCY" };

static void kl_log_meta(klumpat * logger, int level, char * msg);
static void kl_log_file(klumpat * logger, int level, char * msg);
static void kl_log_console(klumpat * logger, int level, char * msg);
static void kl_log_null(klumpat * logger, int level, char * msg);
static void kl_delete_file(klumpat * logger);
static void kl_delete_meta(klumpat * logger);

typedef struct {
  klumpat super;
  char * file;
  int fd;
} klumpat_file;

typedef struct {
  klumpat super;
  klumpat** children;
  unsigned int count;
} klumpat_meta;

#ifdef _WANT_SYSLOG
typedef struct {
  klumpat super;
  int facility;
} klumpat_syslog;
static void kl_log_syslog(klumpat * logger, int leve, char * msg);
#endif

static unsigned int fmt_ulong(char *dest,unsigned long i) {
  register unsigned long len,tmp,len2;
  /* first count the number of bytes needed */
  for (len=1, tmp=i; tmp>9; ++len) tmp/=10;
  if (dest)
    for (tmp=i, dest+=len, len2=len+1; --len2; tmp/=10)
      *--dest = (tmp%10)+'0';
  return len;
}


klumpat * kl_new(char * name, int level) {
  klumpat * self = malloc(sizeof(klumpat));
  if (self) {
    self->type = TYPE_NULL;
    self->name = strdup(name);
    self->level = level;
  } /* if (self) */
  return self;
} /* klumpat * kl_new(char * name, int level) */

klumpat * kl_new_console(char * name, int level) {
  klumpat * self = kl_new(name,level);
  if (self) {
    self->type = TYPE_CONSOLE;
  } /* if (self) */
  return self;
} /* klumpat * kl_new_console(char * name, int level) */


klumpat * kl_new_file(char * name, int level, char * file) {
  klumpat_file * self = malloc(sizeof(klumpat_file));
  if (self) {
    self->super.name = strdup(name);
    self->super.level = level;
    self->super.type = TYPE_FILE;
    self->file = strdup(file);
    self->fd = open(self->file,O_CREAT|O_APPEND|O_WRONLY,0644);
    if (self->fd < 0) {
      kl_delete_file((klumpat *)self);
      self = NULL;
    } /* if (self->fd < 0) */
  } /* if (self) */
  return (klumpat *)self;
} /* klumpat * kl_new_file(char * name, int level, char * file) */

klumpat * kl_new_meta(char * name, int level) {
  klumpat_meta * self = malloc(sizeof(klumpat_meta));
  if (self) {
    self->super.name = strdup(name);
    self->super.level = level;
    self->super.type = TYPE_META;
    self->count = 0;
    self->children = NULL;
  } /* if (self) */
  return (klumpat *)self;
} /* klumpat * kl_new_meta(char * name, int level) */

#ifdef _WANT_SYSLOG
klumpat * kl_new_syslog(char * name, int level, int facility) {
  klumpat_syslog * self = malloc(sizeof(klumpat_syslog));
  if (self) {
    self->super.name = strdup(name);
    self->super.level = level;
    self->super.type = TYPE_SYSLOG;
    self->facility = facility;
  } /* if (self) */
  return (klumpat *)self;
} /* klumpat * kl_new_syslog(char * name, int level, int facility) */
#endif

void kl_add_logger(klumpat * meta, klumpat * log) {
  if (meta->type == TYPE_META) {
    klumpat_meta * k = (klumpat_meta *)meta;
    ++k->count;
    k->children = realloc(k->children,k->count);
    if (k->children) {
      k->children[k->count-1] = log;
    } /* if (k->children) */
  } /* if (klumpat->type == TYPE_META) */
} /* void kl_add_logger(klumpat * meta, klumpat * log) */

void kl_delete(klumpat * logger) {
  switch (logger->type) {
    case TYPE_FILE:
      kl_delete_file(logger);
      break;
    case TYPE_META:
      kl_delete_meta(logger);
      break;
  } /* switch () */
  free(logger->name);
  free(logger);
} /* void kl_delete(klumpat * logger) */

void kl_log(klumpat * logger, int level, char * msg) {
  if (level >= logger->level) {
    switch (logger->type) {
      case TYPE_NULL:
        kl_log_null(logger,level,msg);
        break;
      case TYPE_FILE:
        kl_log_file(logger,level,msg);
        break;
      case TYPE_CONSOLE:
        kl_log_console(logger,level,msg);
        break;
      case TYPE_META:
        kl_log_meta(logger,level,msg);
        break;
#ifdef _WANT_SYSLOG
      case TYPE_SYSLOG:
        kl_log_syslog(logger,level,msg);
        break;
#endif
      default:
        break;
    } /* switch () */
  } /* if (level >= logger->level) */
} /* void kl_log(klumpat * logger, int level, char * msg) */

static void kl_delete_file(klumpat * logger) {
  klumpat_file * k = (klumpat_file *)logger;
  free(k->file);
  close(k->fd);
} /* void kl_delete_file(klumpat * logger) */

static void kl_delete_meta(klumpat * logger) {
  klumpat_meta * k = (klumpat_meta *)logger;
  int i;
  for (i=0;i<k->count;++i) {
    kl_delete(k->children[i]);
  } /* for (;;) */
  free(k->children);
} /* void kl_delete_meta(klumpat * logger) */

static void kl_log_null(klumpat * logger, int level, char * msg) {
  /* nothing */
} /* static kl_log_null(klumpat * logger, int level, char * msg) */

static int select_color(int level) {
  int color = 37;
  switch (level) {
    case INFO: color = 34; break;
    case NOTICE: color = 36; break;
    case WARN: color = 33; break;
    case ERROR: color = 31; break;
    case CRIT:
    case ALERT:
    case EMERG: color = 35; break;
    default:
      break;
  } /* switch () */
  return color;
} /* static int select_color(int level) */

static void set_color(int color) {
  char col_str[10];
  __write1("\033[0;");
  col_str[fmt_ulong(col_str,color)]=0;
  __write1(col_str);
  __write1(";40m");
} /* void set_color(int color) */

static void reset_colors(void) {
  __write1("\033[0m");
} /* void reset_color(void) */

static void kl_log_console(klumpat * logger, int level, char * msg) {
  char * level_name = level_table[level];
  set_color(select_color(level));
  if (level_name) {
    __write1(level_name);
    __write1(" ");
  } /* if (level_name) */
  __write1(logger->name);
  __write1(": ");
  if (msg) {
    __write1(msg);
  } /* if (msg) */
  reset_colors();
  __write1("\n");
} /* static void kl_log_console(klumpat * logger, int level, char * msg) */

static void kl_log_file(klumpat * logger, int level, char * msg) {
  klumpat_file * k = (klumpat_file *)logger;
  char * level_name = level_table[level];
  int fd = k->fd;
  if (level_name) {
    write(fd,level_name,strlen(level_name));
    write(fd," ",1);
  } /* if (level_name) */
  write(fd,logger->name,strlen(logger->name));
  write(fd,": ",2);
  if (msg) {
    write(fd,msg,strlen(msg));
  } /* if (msg) */
  write(fd,"\n",1);
} /* static void kl_log_file(klumpat * logger, int level, char * msg) */

#ifdef _WANT_SYSLOG
static void kl_log_syslog(klumpat * logger, int level, char * msg) {
  klumpat_syslog * k = (klumpat_syslog *)logger;
  openlog(k->super.name,LOG_PID,k->facility);
  syslog(level,msg);
  closelog();
} /* static void kl_log_syslog(klumpat * logger, int level, char * msg) */
#endif

static void kl_log_meta(klumpat * logger, int level, char * msg) {
  klumpat_meta * k = (klumpat_meta *)logger;
  int i;
  for (i=0;i<k->count;++i) {
    kl_log(k->children[i],level,msg);
  } /* for (i=0;i<k->count;++i) */
} /* static void kl_log_meta(klumpat * logger, int level, char * msg) */
